\ 2009-05-17  EW   31_odin/odin.frt
\ demo für Linuxtag 2009

marker --start--

include devices/atmega32.frt    \ register definitions
include lib/bitnames.frt        \ port_pin: high low pin_{out,in}put
: pin_highZ ( pinmask portaddr -- )
  over over pin_input low
;
: pin_pullup_on ( pinmask portaddr -- )
  over over pin_input high
;
\ ----------------------------------------------------------
\ pin layout, bus addresses, IDs ...

\ --- i2c adresses:
$40 constant a_io  \ pcf8574
$a0 constant a_rtc \ pcf8583
$9c constant a_th1 \ lm75
$96 constant a_th2 \ lm92
$62 constant a_sl1 \ sht71 @ atmega32 i2c-slave

\ --- pins:
PORTA 0 portpin: pin_ld         \ adc, air pressure

PORTB 0 portpin: ledtick        \ led timeup ticks
PORTB 1 portpin: ledsec         \ led timeup seconds

PORTC 3 portpin: sht_clk        \ local sht71 sensor
PORTC 4 portpin: sht_dat        \ . needs -jtag in init-loop

\ ----------------------------------------------------------
\ include files ...
include ewlib/dot-res.frt       \ .res ( -- )
include lib/ans94/ans.frt       \ ms nip tuck blank
: pick  ( n -- stack[n] ) cells 2 + sp@ + @ ;
: or!  dup c@ rot or swap c! ;
include ewlib/flags.frt         \ flag: bv fset fclr fset? fclr?
include ewlib/clock_tick.frt    \ tick_isr +ticks -ticks
include lib/twi.frt             \ twi.default twi.ping? twi.scan {+,-}twi
include lib/ans94/2x.frt        \ double cell arithmetic
include ewlib/format.frt        \ sign! d.i .i d+.f +.f
: .id ( id -- )   space .i [ char : ] literal emit ;
include versions.frt             \ values e_stationid e_version
                                \ station version .id.ver banner

include ewlib/timeup.frt              \ timeup: time/date system with jobs
				\ tick sec min day month year
				\ leap_year? length_of_month
				\ tickover? timeup
				\ init-timeup
				\ DT.set  ( Y m d H M S -- )
				\ DT.get  ( -- S M H d m Y )
				\ DT.show ( S M H d m Y -- )
include ewlib/i2c.frt           \ >i2c ( x1 .. xN.msB N addr -- )
                                \ <i2c ( N addr -- xN.msB .. x1 )
include ewlib/i2c_rtc.frt       \ get.rtc ( -- x0 .. x5 yyyy )
                                \ set.rtc ( Y m d H M S S/100 -- )
                                \ show.rtc ( -- )
                                \ init-clock ( -- )
include ewlib/i2c_lmXX.frt      \ lm.get ( i2c_addr -- xh xl ok | err )
                                \ lm75.decode ( xh xl -- T*100 )
                                \ lm92.decode ( xh xl -- T*100 )
include ewlib/sensor.frt        \ sensor:
                                \ create ( id i2c-addr 'get 'decode a_0 -- )
                                \ does>  ( -- id scaled_value ok | id err )
include ewlib/filter_mean.frt   \ filter_mean:
                                \ mean_reset mean_addup mean_eval

include ewlib/portpin.frt       \ pin>channel, needed by adc
include ewlib/adc.frt           \ adc.init, adc.init.pin, adc.get{,10}
include ewlib/pressure_sensor.frt \ pr.get pr.decode

include ewlib/sht75.frt         \ sht.io.init ( -- )        
                                \ sht.reset ( -- )
                                \ sht.get.T ( -- T*100[C] ok | err )
                                \ sht.get.H ( -- H*10[%] ok  | err )

decimal
\ ----------------------------------------------------------
\ i2c slave with sht71 sensor
\ always read 4 byte, drop the unneeded ones
: sl.get.X ( N addr -- x1 .. xN ok | err )
  \ N: number of byte to read from slave
  >r
  r@ twi.ping? if
    0 1 r@ >i2c
    ( N ) r@ <i2c
    0  \ ok
  else
    drop
    -1 \ err
  then
  r> drop
;
: sl.get.T ( addr -- x1 x2 ok | err )
  4 swap sl.get.X 0= if 2drop       0 else -1 then
;
: sl.get.H ( addr -- x1 x2 ok | err )
  4 swap sl.get.X 0= if 2swap 2drop 0 else -1 then
;
: 2bytes>1 ( xl xh -- x ) 8 lshift + ;

\ ----------------------------------------------------------
\ main
\ sensor 1: temperature
1 a_th1 ' lm.get ' lm75.decode    0 sensor: T1
1  filter_mean: F1
\ sensor 2: temperature
2 a_th2 ' lm.get ' lm92.decode    0 sensor: T2
2  filter_mean: F2
\ sensor 3: temperature
\        4: rel humidity
3 a_sl1 ' sl.get.T ' 2bytes>1     0 sensor: T3
3  filter_mean: F3
4 a_sl1 ' sl.get.H ' 2bytes>1     0 sensor: H4
4  filter_mean: F4
\ local mpx4115a sensor
\ sensor 5: pressure
: P5 ( -- id p ok | id err )
  5                       \ id
  pin_ld pr.get pr.decode \ p
  0                       \ ok
;
5 filter_mean: F5
\ local sht71 sensor
\ sensor 6: temperature
\        7: rel humidity
: T6 ( -- id T ok | id err )
  6  sht.get.T
;
6 filter_mean: F6
: H7 ( -- id rH ok | id err )
  7  sht.get.H
;
7 filter_mean: F7
 
\ show result of "Fx mean_eval"
: .F ( id max mean min N>0 | id 0 -- )
  dup if
    5 pick .id            \ id:
    1 .r    [char] , emit \ N,
    2 +.f   [char] , emit \ min,
    2 d+.f  [char] , emit \ d.mean,
    2 +.f                 \ max
    drop \ id
  else
    drop \ N
    .id ." *,*,*,*"
  then
;
: .F2 ( id max mean min N>0 | id 0 -- )
  dup if
    5 pick .id            \ id:
\   1 .r    [char] , emit \ N,
s>d 1 du0.r [ char , ] literal emit \ N,
    1 +.f   [char] , emit \ min,
    1 d+.f  [char] , emit \ d.mean,
    1 +.f                 \ max
    drop \ id
  else
    drop \ N
    .id ." *,*,*,*"
  then
;

: .T ( id value ok | id err -- )
  0= if
    swap .id 2 +.f
  else
    .id ." *"
  then
;
: .P ( id p ok | id err -- )
  0= if
    swap .id 1 +.f
  else
    .id ." *"
  then
;
: .H ( id p ok | id err -- )  .P ;

: init-io
  ledtick high  ledtick pin_output
  ledsec  high  ledsec  pin_output
  timer2 @ cycles/tick + newtimer !
;

: job.tick
  ledtick toggle
  tick @ dup
  0= if ledsec low then
  [ ticks/sec 4 / ] literal = if ledsec high then
;
: job.sec
   sec @ 10 mod 0= if
     \ debug if
     [ char C ] literal emit space
     DT.get DT.show
     space .id.ver
     T1 .T
     T2 .T
     T3 .T
     H4 .H
     P5 .P
     T6 .T
     H7 .H
     cr
     \ then \ end debug if
     T1 0= if F1 mean_addup then drop
     T2 0= if F2 mean_addup then drop
     T3 0= if F3 mean_addup then drop
     H4 0= if F4 mean_addup then drop
     P5 0= if F5 mean_addup then drop
     T6 0= if F6 mean_addup then drop
     H7 0= if F7 mean_addup then drop
   then
   sec @ invert  1 a_io >i2c
;
\ sampling window in minutes
\ must be at least 2 and even!
10 constant t_filter
t_filter 2 / constant m_filter

: job.min
  ." job.min"
  min @ t_filter mod m_filter = if
    space ." 10 minute filter:"
    \ subtract 5 minutes (to middle of sampling window)
    \ if min%10 == 5, then one can savely remove 5
    cr
    [ char F ] literal emit space
    DT.get >r >r >r >r m_filter - r> r> r> r>
    DT.show space .id.ver
    F1 mean_eval .F  F1 mean_reset
    F2 mean_eval .F  F2 mean_reset
    F3 mean_eval .F  F3 mean_reset
    F4 mean_eval .F2 F4 mean_reset
    F5 mean_eval .F2 F5 mean_reset
    F6 mean_eval .F  F6 mean_reset
    F7 mean_eval .F2 F7 mean_reset
  then
  cr
;
: job.hour
  sht.reset
  ." job.hour" cr
;
: job.day
  ." job.day" cr
;
: job.month
  ." job.month" cr
  year @ month @ 1+
  length_of_month
  4 Limits + c!
;
: job.year
  \ fixme: store year in i2c a_rtc @ 0x10
  ." job.year" cr
;

create Jobs
  ' job.tick ,
  ' job.sec , ' job.min ,   ' job.hour ,
  ' job.day , ' job.month , ' job.year ,

variable jobCount
: jobCount++
  jobCount @
  6 < if
    1 jobCount +!
  then
;
: init-loop
  init-io
  init-timeup
  +ticks
  init-twi
  adc.init
  pin_ld adc.init.pin
  -jtag
  sht.io.init
  sht.reset

  a_rtc twi.ping? if
    init-clock
  else
    2000 1 1 0 0 0 DT.set
  then

  1 jobCount !
  F1 mean_reset
  F2 mean_reset
  F3 mean_reset
  F4 mean_reset
  F5 mean_reset
  F6 mean_reset
  F7 mean_reset
;

\ parallel form of timeup:
\ do not call job.sec ... job.year in one go.
\ always call
\     tickover? if timeup ... then
\ after every job.$n
: run-loop
  begin

    \ these are run often
    tickover?
    if
      timeup
      job.tick
      1 jobCount ! \ start with job.sec
    then

    \ these are run one per loop, not all in one go.
    jobCount @
    bv tuFlags fset?
    if
      jobCount @ dup
      Jobs + i@ execute
      bv tuFlags fclr
    then
    jobCount++

  key? until
;

: run
  decimal
  init-loop
  banner
  run-loop
;

: run-turnkey
  baud  +usart  >usart
  +int
  ver
  run
;

\ ' run-turnkey is turnkey
